/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <QtGui/QLabel>
#include <boost/foreach.hpp>

#include <trace-tools/wifi/frame/frame.hh>

#include "wscout_gui_ieee802_11_header.hh"

namespace wscout
{

  namespace gui
  {

    IEEE80211Header::IEEE80211Header(QWidget* parent):
      HeaderBase<IEEE80211Header, protocol::IEEE802_11> (parent)
    {
      setUpdatesEnabled(false);

      setForegroundRole(QPalette::WindowText);
      setBackgroundRole(QPalette::Window);
      setAutoFillBackground(true);

      const QFont	fFixed ("fixed");
      QBoxLayout* const	l = layout();

      {
	QFont bold;
	bold.setBold(true);

	desc.setFont(bold);
	desc.setMinimumWidth(desc.fontMetrics().width('x') * max_desc_len);
	l->addWidget(&desc);
      }

      duration.setMinimumWidth(duration.fontMetrics().width('x') *
			       max_duration_len);
      l->addWidget(&duration);
      l->addSpacing(default_spacing);

      flags.setFont(fFixed);
      l->addWidget(&flags);
      l->addSpacing(default_spacing);

      l->addWidget(&seq_ctl);
      l->addSpacing(before_addresses_spacing);
      l->addStretch();

      {
	for (size_t i = 0; i < sizeof (addr) / sizeof (*addr); ++i)
	  {
	    if (i)
	      {
		addr_sep[i - 1].setFrameStyle(QFrame::VLine | QFrame::Plain);
		l->addWidget(addr_sep + i - 1);
	      }
	    addr[i].setFont(fFixed);
	    l->addWidget(addr + i);
	  }
      }

      setUpdatesEnabled(true);
    }

    IEEE80211Header::next_header
    IEEE80211Header::setup(const void* header, size_t caplen)
    {
      wifi::frame::dissector<ieee80211_hooks>	d (ieee80211_hooks (*this));

      setUpdatesEnabled(false);
      d(header, caplen);
      setUpdatesEnabled(true);
      return next_header (protocol::NONE, 0);
    }

    void
    IEEE80211Header::setColor(const QColor fore, const QColor back)
    {
      QPalette	p;

      p.setColor(QPalette::WindowText, fore);
      p.setColor(QPalette::Window, back);
      setPalette(p);
    }

    IEEE80211Header::ieee80211_hooks::ieee80211_hooks(IEEE80211Header& h):
      h_ (h)
    {
    }

    void
    IEEE80211Header::ieee80211_hooks::frame_hook(const void*, size_t)
    {
      h_.duration.hide();
      h_.flags.hide();
      h_.seq_ctl.hide();
      BOOST_FOREACH(QLabel& l, h_.addr)		l.hide();
      BOOST_FOREACH(QFrame& f, h_.addr_sep)	f.hide();
    }

    void
    IEEE80211Header::ieee80211_hooks::addr_hook(const void*, size_t,
						unsigned i,
						const wifi::addr& a)
    {
      h_.addr[i - 1].setText(make_string(a).c_str());
      h_.addr[i - 1].show();
      if (i - 1)
	h_.addr_sep[i - 2].show();
    }

    void
    IEEE80211Header::ieee80211_hooks::seq_ctl_hook(const void*, size_t,
						   unsigned fragnum,
						   unsigned seqnum)
    {
      h_.seq_ctl.setText(tr("%1-%2").arg(fragnum).arg(seqnum));
      h_.seq_ctl.show();
    }

    void
    IEEE80211Header::
    ieee80211_hooks::management_hook(const wifi::frame::mgt::header* h, size_t)
    {
      static const attributes	subtypes[16] =
	{
	  { "Assoc. request (MGT)",		black, lightblue },
	  { "Assoc. response (MGT)",		black, lightblue },
	  { "Reassoc. request (MGT)",		black, lightblue },
	  { "Reassoc. response (MGT)",		black, lightblue },
	  { "Probe request (MGT)",		black, lightblue },
	  { "Probe response (MGT)",		black, lightblue },
	  { "Invalid (MGT:6)",			lightblue, red },
	  { "Invalid (MGT:7)",			lightblue, red },
	  { "Beacon (MGT)",			black, lightblue },
	  { "Ann. traffic indic. mess. (MGT)",	black, lightblue },
	  { "Disassoc. (MGT)",			black, lightblue },
	  { "Auth. (MGT)",			black, lightblue },
	  { "Deauth. (MGT)",			black, lightblue },
	  { "Invalid (MGT:13)",			lightblue, red },
	  { "Invalid (MGT:14)",			lightblue, red },
	  { "Invalid (MGT:15)",			lightblue, red },
	};

      setDesc(subtypes[wifi::frame::subtype_of(h)]);
      setControl(h);
    }

    void
    IEEE80211Header::ieee80211_hooks::control_hook(const void* h, size_t)
    {
      static const attributes	subtypes[16] =
	{
	  { "Invalid (CTL:0)",	yellow, red },
	  { "Invalid (CTL:1)",	yellow, red },
	  { "Invalid (CTL:2)",	yellow, red },
	  { "Invalid (CTL:3)",	yellow, red},
	  { "Invalid (CTL:4)",	yellow, red },
	  { "Invalid (CTL:5)",	yellow, red },
	  { "Invalid (CTL:6)",	yellow, red },
	  { "Invalid (CTL:7)",	yellow, red },
	  { "Invalid (CTL:8)",	yellow, red },
	  { "Invalid (CTL:9)",	yellow, red },
	  { "PS-Poll (CTL)",	black, yellow },
	  { "RTS (CTL)",	black, yellow },
	  { "CTS (CTL)",	black, yellow },
	  { "Ack. (CTL)",	black, yellow },
	  { "CF-End (CTL)",	black, yellow },
	  { "CF-Ack (CTL)",	black, yellow },
	};

      setDesc(subtypes[wifi::frame::subtype_of(h)]);
      setControl(h);
    }

    void
    IEEE80211Header::ieee80211_hooks::data_hook(const void* h, size_t)
    {
      static const attributes	subtypes[16] =
	{
	  { "Data",			black, lightgreen },
	  { "Data + CF-Ack",		black, lightgreen },
	  { "Data + CF-Poll",		black, lightgreen },
	  { "Data + CF-Ack + CF-Poll",	black, lightgreen },
	  { "Null (no data)",		black, lightgreen },
	  { "CF-Ack (DATA)",		black, lightgreen },
	  { "CF-Poll (DATA)",		black, lightgreen },
	  { "CF-Ack + CF-Poll (DATA)",	black, lightgreen },
	  { "Invalid (DATA:8)",		lightgreen, red },
	  { "Invalid (DATA:9)",		lightgreen, red },
	  { "Invalid (DATA:10)",	lightgreen, red },
	  { "Invalid (DATA:11)",	lightgreen, red },
	  { "Invalid (DATA:12)",	lightgreen, red },
	  { "Invalid (DATA:13)",	lightgreen, red },
	  { "Invalid (DATA:14)",	lightgreen, red },
	  { "Invalid (DATA:15)",	lightgreen, red },
	};

      setDesc(subtypes[wifi::frame::subtype_of(h)]);
      setControl(h);
    }

    void
    IEEE80211Header::ieee80211_hooks::invalid_type_hook(const void* h, size_t)
    {
      const unsigned	type = wifi::frame::type_of(h);
      const unsigned	subtype = wifi::frame::subtype_of(h);

      h_.setColor(white, red);
      h_.desc.setText(tr("Invalid frame type (%1:%2)").arg(type).arg(subtype));
    }

    void
    IEEE80211Header::ieee80211_hooks::
    truncated_frame_hook(const void*, size_t,
			 wifi::frame::dissector_status::status s)
    {
      static const attributes truncated = { "Truncated frame", white, red };

      if (s == wifi::frame::dissector_status::invalid)
	setDesc(truncated);
    }

    void
    IEEE80211Header::ieee80211_hooks::setDesc(const attributes& attr)
    {
      h_.setColor(attr.fore, attr.back);
      h_.desc.setText(tr(attr.desc.c_str()));
    }

    void
    IEEE80211Header::ieee80211_hooks::setControl(const void* f)
    {
      // Duration
      h_.duration.setText(tr("duration: %1").arg(wifi::frame::duration_of(f)));
      h_.duration.show();

      // Flags.
      {
	static const char	set[] = "TFGRPDWO";
	char			res[sizeof (set)];

	for (size_t i = 0; i < sizeof (set) - 1; ++i)
	  res[i] = wifi::frame::control_flag(i, f) ? set[i] : '-';
	res[sizeof (set) - 1] = '\0';
	h_.flags.setText(res);
	h_.flags.show();
      }
    }

    const QColor IEEE80211Header::ieee80211_hooks::black ("black");
    const QColor IEEE80211Header::ieee80211_hooks::lightblue ("lightblue");
    const QColor IEEE80211Header::ieee80211_hooks::lightgreen ("lightgreen");
    const QColor IEEE80211Header::ieee80211_hooks::red ("red");
    const QColor IEEE80211Header::ieee80211_hooks::white ("white");
    const QColor IEEE80211Header::ieee80211_hooks::yellow ("yellow");

  } // End of namespace gui.

} // End of namespace wscout.
