/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <QtGui/QVBoxLayout>

#include <boost/foreach.hpp>

#include "wscout_gui_packet.hh"

namespace wscout
{

  namespace gui
  {

    Packet::Packet(QWidget* parent): QFrame (parent)
    {
      setFrameStyle(QFrame::Box | QFrame::Plain);
      setSizePolicy(QSizePolicy (QSizePolicy::Minimum, QSizePolicy::Fixed));
      setupLayout();
      setSelected(false);
    }

    void
    Packet::setup(link_type t, unsigned n, const packet& p)
    {
      setUpdatesEnabled(false);

      BOOST_FOREACH(AbstractHeader* h, headers_)
	delete h;
      headers_.clear();

      AbstractHeader::next_header	next = pcap_.setup(t, n, p);
      while (true)
	{
	  const uint8_t* const	b = static_cast<const uint8_t*> (next.second);
	  const uint8_t* const	e = p.data.get();
	  const size_t		caplen = p.caplen - (b - e);

	  protocol_map&			m = protocols();
	  protocol_map::iterator	p = m.find(next.first);

	  if (p == m.end())
	    break;
	  AbstractHeader* h = p->second();
	  headers_.push_back(h);
	  next = h->setup(next.second, caplen);
	  layout()->addWidget(h);
	  h->show();
	}

      updateMinimumSize();
      setUpdatesEnabled(true);
    }

    QBoxLayout*
    Packet::layout()
    {
      QLayout* const	l = QFrame::layout();
      QBoxLayout* const	r = dynamic_cast<QBoxLayout*> (l);

      assert(not l or r);
      return r;
    }

    void
    Packet::registerProtocol(protocol::id id, header_allocator a)
    {
      protocols()[id] = a;
    }

    Packet::protocol_map&
    Packet::protocols()
    {
      static protocol_map	m;

      return m;
    }

    void
    Packet::setSelected(bool state)
    {
      selected_ = state;
      setupBorders();
    }

    int
    Packet::lineWidth() const
    {
      return selected_ ? border_sel : border_unsel;
    }

    int
    Packet::margin() const
    {
      return selected_ ? packet_margin_sel : packet_margin_unsel;
    }

    void
    Packet::setupBorders()
    {
      const int m = margin();

      setUpdatesEnabled(false);

      setLineWidth(lineWidth());
      layout()->setContentsMargins(m, m, m, m);
      updateMinimumSize();

      setUpdatesEnabled(true);
    }

    void
    Packet::setupLayout()
    {
      QVBoxLayout* l = new QVBoxLayout;

      l->setSpacing(header_spacing);
      l->addWidget(&pcap_);
      setLayout(l);
    }

    void
    Packet::updateMinimumSize()
    {
      const int l = lineWidth();

      setMinimumSize(layout()->minimumSize() + QSize (l, l) * 2);
    }

  } // End of namespace gui.

} // End of namespace wscout.
