/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#ifndef WSCOUT_GUI_TRACE_HH_
# define WSCOUT_GUI_TRACE_HH_

# include <ostream>

# include <QtGui/QAbstractScrollArea>
# include <QtCore/QTimer>

# include "wscout_gui_packet.hh"
# include "wscout_pcap_trace.hh"

namespace wscout
{

  namespace gui
  {

    /// Scrollable list of packets.
    class Trace: public QAbstractScrollArea
    {
      Q_OBJECT

      /// Position in the underlying pcap::Trace instance.
      Q_PROPERTY(int position
		 READ position
		 WRITE setPosition
		 RESET setPosition
		 DESIGNABLE false
		 NOTIFY positionChanged(int)
		 USER true)

      enum
      {
	/// Delay before updating after a call to setPosition().
	set_position_update_delay		= 0,

	/// Delay before updating after the scrollbar has been moved.
	scroll_contents_by_update_delay		= 20,

	/// Space between packets.
	packet_spacing				= 4,

	/// Vertical margins (top and bottom).
	trace_vmargin				= packet_spacing,

	/// Horizontal margins (left and right).
	trace_hmargin				= 8
      };

    public:
      /// Build a trace from a pcap::Trace instance.
      explicit Trace(pcap::Trace* trace = 0, QWidget* parent = 0);
      virtual ~Trace();

      /// Change the underlying pcap::Trace instance.
      void			setTrace(pcap::Trace* trace = 0);

      /// Retrieve the underlying pcap::Trace instance.
      const pcap::Trace*	trace() const;

      /// Current position in the underlying pcap::Trace instance.
      int		position() const;

      /// Last position in the underlying pcap::Trace instance.
      int		lastPosition() const;

      /**
       * Get a summary of a given position as a string.
       *
       * This summary consists of the current position, the last
       * position, and the ratio of the two.
       *
       * @e e.g. @code 42 / 84 (50%) @endcode
       *
       * @param[in]	pos	The position to summarize.  By default the
       *			current position.
       */
      QString		positionString(int pos = -1) const;

    signals:
      /// The underlying pcap::Trace instance changed.
      void		traceChanged();

      /// The current position changed.
      void		positionChanged(int);

    public slots:
      /// Change the current position.
      void		setPosition(int p = 0);

      /// Update the widget's content.
      virtual void	update();

    protected:
      /// @name Qt overrided functions
      //@{
      virtual void	keyPressEvent(QKeyEvent*);
      virtual void	mouseMoveEvent(QMouseEvent*);
      virtual void	mousePressEvent(QMouseEvent*);
      virtual void	resizeEvent(QResizeEvent*);
      virtual void	scrollContentsBy(int, int);
      //@}

    private:
      int		packetAt(const QPoint& pos) const;
      void		startUpdateTimer(int ms);
      void		resizeView(const QSize&	size,
				   bool		resize_pkts,
				   bool		update_if_needed);

      struct selection
      {
	struct range: public std::pair<int, int>
	{
	  range(int, int);
	};

	void		clear();
	bool		contains(int) const;
	std::ostream&	print(std::ostream&) const;
	selection&	toggle(int);

	selection&	operator = (int);
	selection&	operator = (const range&);

      private:
	typedef std::list<range>		list_of_ranges;
	typedef list_of_ranges::iterator	iterator;

	void		cleanup(iterator&);

	list_of_ranges	ranges_;
      };

      pcap::Trace*		trace_;
      QWidget			view_;
      int			pos_;
      QTimer			update_timer_;
      std::vector<Packet*>	pkts_;
      selection			selection_;
      int			last_selected_;
    };

  } // End of namespace gui.

} // End of namespace wscout.

#endif // ! WSCOUT_GUI_TRACE_HH_
