/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef PCAP_DESCRIPTOR_HXX_
# define PCAP_DESCRIPTOR_HXX_

# include <stdexcept>

# include "descriptor.hh"

namespace pcapxx
{

  template <class B>
  descriptor<B>::descriptor(const std::string&	dev,
			    int			snaplen,
			    int			promisc,
			    int			to_ms)
  {
    char errbuf[PCAP_ERRBUF_SIZE];

    // FIXME: Why does libpcap need a *non-const* char *dev?
    d_.reset(new shared_data (pcap_open_live(const_cast<char*> (dev.c_str()),
					     snaplen, promisc, to_ms, errbuf)));
    if (not d_->desc)
      throw std::invalid_argument (errbuf);
  }

  template <class B>
  descriptor<B>::descriptor(const std::string& fname)
  {
    char errbuf[PCAP_ERRBUF_SIZE];

    d_.reset(new shared_data (pcap_open_offline(fname.c_str(), errbuf)));
    if (not d_->desc)
      throw std::invalid_argument (errbuf);
  }

  template <class B>
  descriptor<B>::descriptor(int linktype, int snaplen)
  {
    d_.reset(new shared_data (pcap_open_dead(linktype, snaplen)));
    if (not d_->desc)
      throw std::runtime_error ("unknown libpcap error");
  }

  template <class B>
  int
  descriptor<B>::datalink() const
  {
    return pcap_datalink(d_->desc);
  }

  template <class B>
  bool
  descriptor<B>::is_swapped() const
  {
    return pcap_is_swapped(d_->desc);
  }

  template <class B>
  std::string
  descriptor<B>::error() const
  {
    return pcap_geterr(d_->desc);
  }

  template <class B>
  pcapxx::dumper
  descriptor<B>::dumper(const std::string& filename) const
  {
    if (pcap_dumper_t* r = pcap_dump_open(d_->desc, filename.c_str()))
      return pcapxx::dumper (r);
    else
      throw std::runtime_error (error());
  }

  template <class B>
  template <handler Callback>
  int
  descriptor<B>::loop(void* user, int cnt)
  {
    wrapper_data d;

    d.desc = this;
    d.user = user;

    int r = pcap_loop(d_->desc, cnt,
		      handler_wrapper<Callback>,
		      reinterpret_cast<u_char*> (&d));

    if (r == -1)
      throw std::runtime_error (error());
    return r;
  }

  template <class B>
  descriptor<B>::shared_data::shared_data(pcap_t* d): desc (d), curr (0)
  {
  }

  template <class B>
  descriptor<B>::shared_data::~shared_data()
  {
    pcap_close(desc);
  }

  template <class B>
  template <handler H>
  void
  descriptor<B>::handler_wrapper(u_char*		user,
				 const pcap_pkthdr*	h,
				 const u_char*		bytes)
  {
    wrapper_data&	d = *reinterpret_cast<wrapper_data*> (user);
    pkthdr		h_;

    h_.id = ++(d.desc->d_->curr);
    h_.ts = h->ts;
    h_.caplen = h->caplen;
    h_.len = h->len;

    return H(d.user, &h_, bytes);
  }

} // End of namespace pcapxx.

#endif // ! PCAP_DESCRIPTOR_HXX_
