/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_FILTER_INTERSECTOR_HXX_
# define WIFI_FRAME_FILTER_INTERSECTOR_HXX_

# include "intersector.hh"

namespace wifi
{
  namespace frame
  {
    namespace filter
    {

      namespace internals
      {

	template <class I1, class I2, class B1, class B2>
	intersector_iterator<I1, I2, B1, B2>::
	intersector_iterator(iterable_type& i, bool end):
	  tool::iterator_base<exact_type> (i, end)
	{
	}

	template <class I1, class I2, class B1, class B2>
	void
	intersector_iterator<I1, I2, B1, B2>::increment()
	{
	  arrivals1_type&	arrivals1 = this->i_->arrivals1_;
	  arrivals2_type&	arrivals2 = this->i_->arrivals2_;
	  I1&			next1 = this->i_->next1_;
	  I2&			next2 = this->i_->next2_;
	  const I1&		last1 = this->i_->last1_;
	  const I2&		last2 = this->i_->last2_;

	  while (next1 != last1 or next2 != last2)
	    if ((next1 != last1 and check<false>(next1, arrivals1, arrivals2))
		or
		(next2 != last2 and check<true>(next2, arrivals2, arrivals1)))
	      return;
	  arrivals1.clear();
	  arrivals2.clear();
	  this->v_ = boost::none_t ();
	}

	template <class I1, class I2, class B1, class B2>
	template <bool SwapResult, class ArrType, class RHSArrType>
	bool
	intersector_iterator<I1, I2, B1, B2>::
	check(I1&		frm,
	      ArrType&		frm_arrivals,
	      RHSArrType&	rhs_arrivals)
	{
	  typename RHSArrType::iterator i = rhs_arrivals.find(*frm);

	  if (i == rhs_arrivals.end())
	    {
	      frm_arrivals.insert(*frm);
	      ++frm;
	      return false;
	    }

	  this->v_ = SwapResult ?
	    std::make_pair(*i, *frm):
	    std::make_pair(*frm, *i);

	  // Cleanup the set of arrivals (i.e. remove frames previous
	  // to this one).  We should do it systematically, but that
	  // operation takes time.  So we ignore it by default.  We
	  // only do it if the set of arrivals is too big (i.e. there
	  // is a risk of saturating memory, or the search time may
	  // become non-negligible).
	  {
	    enum { arrivals_max_size = 0 }; // 1000000 };

	    if (rhs_arrivals.size() > arrivals_max_size)
	      cleanup_arrivals(i, rhs_arrivals);
	  }

	  rhs_arrivals.erase(i);
	  ++frm;
	  return true;
	}

	template <class I1, class I2, class B1, class B2>
	template <class ArrType>
	void
	intersector_iterator<I1, I2, B1, B2>::
	cleanup_arrivals(const typename ArrType::iterator&	frame,
			 ArrType&				arrivals)
	{
	  const tool::microseconds	m = frame->microseconds();
	  typename ArrType::iterator	i = arrivals.begin();

	  while (i != arrivals.end())
	    {
	      if (i->microseconds() < m)
		arrivals.erase(i++);
	      else
		++i;
	    }
	}

      } // End of namespace internals.


      template <class I1, class I2, class B>
      intersector<I1, I2, B>::intersector(const I1& first1, const I1& last1,
					  const I2& first2, const I2& last2):
	arrivals1_ (),
	arrivals2_ (),
	next1_ (first1),
	last1_ (last1),
	next2_ (first2),
	last2_ (last2)
      {
      }

    } // End of namespace wifi::frame::filter.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

#endif // ! WIFI_FRAME_FILTER_INTERSECTOR_HXX_
