/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_FILTER_UNIQUELY_IDENTIFIABLE_HH_
# define WIFI_FRAME_FILTER_UNIQUELY_IDENTIFIABLE_HH_

# include <trace-tools/prism_header.hh>
# include <trace-tools/tool/iterator_base.hh>
# include <trace-tools/tool/iterable.hh>
# include <trace-tools/wifi/addr_mapping.hh>
# include <trace-tools/wifi/mactime_tracker.hh>

# include "uniquely_identifiable_fwd.hh"

namespace tool
{

  /**
   * Virtual %types for wifi::frame::filter::internals::ui_iterator.
   * @see types
   */
  template <class U, class I, class H, class B1, class B2>
  struct types< wifi::frame::filter::internals::ui_iterator<U, I, H, B1, B2> >
  {
    typedef
      wifi::frame::filter::internals::ui_iterator_value_type<U>
      value_type;

    typedef
      wifi::frame::filter::uniquely_identifiable<U, I, H, B1>
      iterable_type;
  };

  /**
   * Virtual %types for wifi::frame::filter::uniquely_identifiable.
   * @see types
   */
  template <class U, class I, class H, class B>
  struct types< wifi::frame::filter::uniquely_identifiable<U, I, H, B> >
  {
    typedef
      wifi::frame::filter::internals::ui_iterator<U, I, H, B, bottom>
      iterator;
  };

} // End of namespace tool.

namespace wifi
{
  namespace frame
  {

    /**
     * Various %frame filters.
     *
     * A %filter is an iterable object which filters out various
     * elements from a given iterator.
     */
    namespace filter
    {

      /// Internal compounds - not designed for users.
      namespace internals
      {

	/// @link ui_iterator @endlink's value %type.
	template <class UniqueId>
	struct ui_iterator_value_type: public UniqueId
	{
	  /**
	   * Construct from a microsecond count and a uniquely identifiable
	   * %frame.
	   */
	  ui_iterator_value_type(const UniqueId&,
				 const tool::microseconds&);

	  /// Frame's time of arrival in microseconds.
	  const tool::microseconds&	microseconds() const;

	private:
	  tool::microseconds		microseconds_;
	};

	/// Iterator over uniquely identifiable frames.
	template <class U, class I, class H, class B, class Bottom>
	struct ui_iterator: tt_inherit(public tool::iterator_base,
				       ui_iterator<U, I, H, B, Bottom>)
	{
	  /// This template instantiation's exact %type.
	  typedef
	    tt_get_exact(Bottom, ui_iterator<U, I, H, B, Bottom>) exact_type;
	  typedef tt_type(iterable_type, exact_type)		  iterable_type;

	  /// @link tool::iterable @endlink-compatible constructor.
	  ui_iterator(iterable_type& uif, bool end);

	  /// Retrieve the next uniquely identifiable %frame.
	  void	increment();
	};

      } // End of namespace wifi::frame::filter::internals.


      /// Filter uniquely identifiable frames out of PCAP frames.
      template <class UniqueId,
		class InputIterator,
		class HeaderType = prism::header,
		class Bottom = tool::bottom>
      struct uniquely_identifiable:
	tt_inherit(public tool::iterable,
		   uniquely_identifiable<UniqueId, InputIterator, HeaderType,
					 Bottom>)
      {
	/**
	 * Construct from two iterators and a @link addr_mapping
	 * address mapping @endlink.
	 */
	uniquely_identifiable(const InputIterator& begin,
			      const InputIterator& end,
			      addr_mapping&);

	bool		prism_header_noise_means_error;

      private:
	InputIterator	next_;
	InputIterator	end_;
	addr_mapping*	mapping_;
	mactime_tracker	mactime_;

	friend class internals::ui_iterator<UniqueId,
					    InputIterator,
					    HeaderType,
					    Bottom, tool::bottom>;
      };

    } // End of namespace wifi::frame::filter.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

# include "uniquely_identifiable.hxx"

#endif // ! WIFI_FRAME_FILTER_UNIQUELY_IDENTIFIABLE_HH_
