/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_UNIQUE_ID_HH_
# define WIFI_FRAME_UNIQUE_ID_HH_

# include <ostream>
# include <boost/optional.hpp>

# include <trace-tools/pcap/frame_descriptor.hh>
# include <trace-tools/wifi/addr_mapping.hh>

namespace wifi
{

  namespace frame
  {

    /// Internal compounds - not designed for users.
    namespace internals
    {
      struct tag1 {};
      struct tag2 {};
      struct tag3 {};
      struct tag4 {};

      template <class Tag>
      struct empty
      {
	bool operator  < (const empty& rhs) const;
	bool operator == (const empty& rhs) const;
	bool operator != (const empty& rhs) const;
      };

      template <class T, class Tag = tag1>
      struct adapter
      {
	adapter(const T& v = T());

	bool operator  < (const adapter& rhs) const;
	bool operator == (const adapter& rhs) const;
	bool operator != (const adapter& rhs) const;

      private:
	T	v_;
      };

    } // End of namespace wifi::frame::internals.

    /**
     * Identifier for uniquely identifiable frames.
     *
     * A uniquely identifiable %frame is a %frame that is unique
     * inside a given trace.
     *
     * @todo Document.
     */
    template < class Id1 = internals::empty<internals::tag1>,
	       class Id2 = internals::empty<internals::tag2>,
	       class Id3 = internals::empty<internals::tag3>,
	       class Id4 = internals::empty<internals::tag4> >
    struct unique_id: private Id1, private Id2, private Id3, private Id4
    {
      /**
       * Construct a uniquely identifiable %frame from a %frame id
       * (from its @link pcapxx::pkthdr PCAP header @endlink) and some
       * distinctive attributes.
       *
       * The %frame id's only role is to help debugging.  It is only
       * used when dumping the %frame to a C++ stream.
       *
       * @see pcapxx::frame_descriptor
       */
      unique_id(unsigned	frame_id,
		const Id1&	id1 = Id1 (),
		const Id2&	id2 = Id2 (),
		const Id3&	id3 = Id3 (),
		const Id4&	id4 = Id4 ());

      /// @name Comparison operators
      //@{
      bool		operator < (const unique_id& rhs) const;
      bool		operator == (const unique_id& rhs) const;
      bool		operator != (const unique_id& rhs) const;
      //@}

      /// Dump a uniquely identifiable %frame onto a C++ stream.
      std::ostream&	print(std::ostream&) const;

      /// Get the frame id (for debugging purposse).
      unsigned		frame_id() const;

      /// @name Access to the distinctive attributes.
      //@{
      const Id1&	id1() const;
      const Id2&	id2() const;
      const Id3&	id3() const;
      const Id4&	id4() const;
      //@}

    private:
      unsigned		frame_id_;
    };

    /// Dump a @link uniquely_identifiable uniquely identifiable
    /// %frame @endlink onto a C++ stream.
    template <class Id1, class Id2, class Id3, class Id4>
    std::ostream&
    operator << (std::ostream&, const unique_id<Id1, Id2, Id3, Id4>&);

    /**
     * Make a unique id of a %frame (if possible).
     *
     * @warning This function is only declared (and @b not defined) so
     *		the compiler knows other make_unique_id()'s are template
     *		specializations.  Calling it will not link.
     */
    template <class UniqueId, class HeaderType>
    boost::optional<UniqueId>
    make_unique_id(const pcapxx::frame_descriptor& desc,
		   addr_mapping&		mapping,
		   bool				prism_noise_means_error);

    namespace internals
    {

      bool
      make_unique_id_handle_prism(const void*&	frame,
				  size_t&	caplen,
				  const pcapxx::frame_descriptor& desc,
				  bool		noise_means_error);

    } // End of namespace wifi::frame::internals.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

# include "unique_id.hxx"

#endif // ! WIFI_FRAME_UNIQUE_ID_HH_
