/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#include <iostream>
#include <set>
#include <map>
#include <stdexcept>
#include <cstdlib>

#include <boost/format.hpp>

#include "config.hh"

#include <trace-tools/pcap/descriptor.hh>
#include <trace-tools/wifi/addr_mapping.hh>
#include <trace-tools/wifi/frame/source_timestamp_id.hh>
#include <trace-tools/wifi/frame/bssid_timestamp_id.hh>
#include <trace-tools/wifi/frame/source_bssid_timestamp_id.hh>
#include <trace-tools/wifi/frame/filter/uniquely_identifiable.hh>

namespace
{

  struct collision
  {
    collision(const unsigned rhs_id, const tool::microseconds& interval):
      rhs_id (rhs_id),
      interval (interval)
    {
    }

    unsigned		rhs_id;
    tool::microseconds	interval;
  };

  template <class Filter, class InputIterator>
  void
  test_uniqueness(const InputIterator& begin, const InputIterator& end,
		  wifi::addr_mapping& m)
  {
    typedef typename Filter::iterator			filter_iterator;
    typedef typename filter_iterator::value_type	arrival;
    typedef std::set<arrival>				set_of_arrivals;
    typedef std::vector<collision>			collisions_vector;
    typedef std::map<arrival, collisions_vector>	collisions_map;
    typedef typename collisions_vector::const_iterator	cvec_iterator;
    typedef typename collisions_map::const_iterator	cmap_iterator;

    Filter		filter (begin, end, m);
    set_of_arrivals	arrivals;
    collisions_map	collisions;
    unsigned		arrival_count = 0;
    unsigned		collision_count = 0;

    // FIXME: Pass as a program argument.
    filter.prism_header_noise_means_error = true;

    for (filter_iterator i = filter.begin(); i != filter.end(); ++i)
      {
	typedef typename set_of_arrivals::iterator iterator;

	const std::pair<iterator, bool> r = arrivals.insert(*i);

	++arrival_count;
	if (not r.second)
	  {
	    const tool::microseconds dt (i->microseconds() -
					 r.first->microseconds());

	    ++collision_count;
	    collisions[*r.first].push_back(collision (i->frame_id(), dt));
	  }
      }

    boost::format f ("%15s %15s %15s");

    std::cout << (f % "frame 1" % "frame 2" % "interval (us)") << '\n';
    std::cout << (f % "=======" % "=======" % "=============") << std::endl;
    for (cmap_iterator i = collisions.begin(); i != collisions.end(); ++i)
      {
	bool first = true;
	for (cvec_iterator j = i->second.begin(); j != i->second.end(); ++j)
	  {
	    std::cout << (f % i->first % j->rhs_id % j->interval);
	    if (first)
	      {
		first = false;
		std::cout << " (new)";
	      }
	    std::cout << '\n';
	  }
      }
    std::cerr << arrival_count << " unique frames, "
	      << collision_count << " collisions." << std::endl;
  }

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  if (argc != 3)
    {
      std::cerr << "Usage:\n\t"  << argv[0]
		<< " <distinctive_attributes> <mapping_file>" << std::endl;
      return EXIT_FAILURE;
    }

  try
    {
      using namespace wifi::frame;

      typedef pcapxx::descriptor<>	pcap_descriptor;

      typedef filter::uniquely_identifiable<source_timestamp_id,
					    pcap_descriptor::iterator>
	      st_uniquely_identifiable;

      typedef filter::uniquely_identifiable<bssid_timestamp_id,
					    pcap_descriptor::iterator>
	      bt_uniquely_identifiable;

      typedef filter::uniquely_identifiable<source_bssid_timestamp_id,
					    pcap_descriptor::iterator>
	      sbt_uniquely_identifiable;

      wifi::addr_mapping	m (argv[2]);
      pcapxx::descriptor<>	d ("-");
      if (std::string (argv[1]) == "source-timestamp")
	test_uniqueness<st_uniquely_identifiable>(d.begin(), d.end(), m);
      else if (std::string(argv[1]) == "bssid-timestamp")
	test_uniqueness<bt_uniquely_identifiable>(d.begin(), d.end(), m);
      else if (std::string(argv[1]) == "source-bssid-timestamp")
	test_uniqueness<sbt_uniquely_identifiable>(d.begin(), d.end(), m);
      else
	throw std::invalid_argument ("Invalid distinctive attributes");
    }
  catch (const std::exception& e)
    {
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
      return EXIT_FAILURE;
    }
}
