/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIT_IMPORTER_HH_
# define WIT_IMPORTER_HH_

# include <string>
# include <map>
# include <vector>

# include <boost/date_time/posix_time/posix_time.hpp>
# include <boost/shared_ptr.hpp>

# include <trace-tools/tool/datafile.hh>
# include <trace-tools/pcap/descriptor.hh>
# include <trace-tools/wifi/addr.hh>
# include <trace-tools/wifi/addr_mapping.hh>
# include <trace-tools/prism_header.hh>
# include <trace-tools/wifi/frame/dissector/dissector.hh>
# include <trace-tools/wifi/frame/dissector/default_hooks.hh>

namespace wit
{

  /**
   * Functor to create WIT data files from PCAP traces.
   *
   * The created data files are designed to be loaded into databases
   * using LOAD DATA INFILE SQL commands.
   */
  struct importer
  {
    /// WIT's createRawDataDatabase.pl error codes.
    enum error
      {
	success			= 0,
	malformed_frame		= 1,
	malformed_prism		= 2,
	malformed_802		= 3,
	malformed_other		= 4,
	driver_reported		= 5,
	unexpected_section	= 6,
	bad_length		= 7,
	bad_type_subtype 	= 8
      };

    /**
     * Contruct a wit::importer given a prefix to use for data files.
     *
     * The created data files will have the name @c
     * <prefix>_c<channel> (for raw frames) and @c
     * <prefix>_c<channel>_MACIndex (for address to integer mappings).
     */
    importer(const std::string& prefix);

    /**
     * Import a frame in the data files.
     *
     * @param[in] h	Pointer to the frame's PCAP header.
     * @param[in] p	Pointer to the frame.
     */
    void	got_frame(const pcapxx::pkthdr* h, const void* p);

  protected:
    /// NULL field representation.
    static const std::string	db_null;

    enum
      {
	/// Maximum number of Wi-Fi channels.
	channels_max		= 14
      };

    /*----------------.
    | importer::hooks |
    `----------------*/

    /// Parse a 802.11 frame and retrieve some fields that wit::importer needs.
    struct hooks: public wifi::frame::dissector_default_hooks
    {
      hooks(wifi::addr_mapping& m);

      std::string	addrs() const;
      std::string	beacon_timestamp() const;
      std::string	fragnum() const;
      std::string	seqnum() const;
      std::string	error() const;

    protected:

      void	invalid_type_or_subtype_hook(const void*, size_t, status);
      void	truncated_frame_hook(const void*, size_t, status);
      void	addr_hook(const void*, size_t, unsigned, const wifi::addr&);
      void	seq_ctl_hook(const void*, size_t, unsigned, unsigned);
      void	beacon_hook(const wifi::frame::mgt::header*, size_t);
      void	end_of_frame_hook(const void*, size_t, status);

    private:

      wifi::addr_mapping&	mapping_;
      std::string		addrs_;
      std::string		beacon_timestamp_;
      std::string		fragnum_;
      std::string		seqnum_;
      std::string		error_;
    };


    /// Get the name of a frames data file given a prefix and a channel.
    static std::string	frames_data_filename(const std::string&, unsigned);

    /// Get the name of a mapping data file given a prefix and a channel.
    static std::string	mapping_filename(const std::string&,  unsigned);

    /**
     * Update the object's time information and return a time string
     * for a given frame's PCAP header.
     */
    std::string		handle_time_data(const pcapxx::pkthdr*);

    /// Update data files when a Prism header was trucated.
    void		handle_malformed_prism();

    /**
     * Parse a frame's Prism header and update data files accordingly.
     *
     * @return	A pointer to the Prism header, or NULL if the header is
     *		truncated.  In this case, the data files have been
     *		fully updated and processing may stop safely for this
     *		frame.
     */
    const prism::header*
    handle_prism_header(const std::string& time_string,
			const void*,
			size_t caplen);

    /// Channel of last received frame.
    unsigned				last_channel_;

    /// Time of arrival of last received frame.
    boost::posix_time::ptime		last_frame_;

    /// ID of last receveid frame (per channel).
    unsigned				last_id_[channels_max];

    /// Frames data for each channel.
    std::vector<tool::datafile>		frames_data_;

    /// Address to index mapping for each channel.
    std::vector<wifi::addr_mapping>	mappings_;
  };

} // End of namespace wit.

# include "wit_importer.hxx"

#endif // ! WIT_IMPORTER_HH_
